"""library.py - Library Dialog.

Copyright (C) 2011-2017 Robert Pastierovic
https://launchpad.net/~just-me
"""

# -------------------------------------------------------------------------
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# -------------------------------------------------------------------------


import gtk
import pygtk
import os.path
import lxml.etree as xml
import zipfile
import shutil
from PIL import Image
import cStringIO
import base64
from xml.sax.saxutils import escape, unescape
from random import randint

import constants
import filechooser
import fileprepare
import acbfdocument
import library_info
import preferences

class LibraryDialog(gtk.Dialog):
    
    """Library dialog."""
    
    def __init__(self, window):
        self._window = window
        self.library = Library()
        gtk.Dialog.__init__(self, 'Comic Books Library', self, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT)
        self.set_resizable(True)
        self.set_border_width(8)
        self.ui_scale_factor = self._window.ui_scale_factor
        self.set_geometry_hints(min_height=360 * self._window.ui_scale_factor, min_width=635 * self._window.ui_scale_factor)
        self.set_size_request(900 * self._window.ui_scale_factor, 600 * self._window.ui_scale_factor)
        self.isChanged = False
        self.preferences = preferences.Preferences()

        self.connect('key_press_event', self.key_pressed)
        self.connect('check-resize', self.resize_library)
        self.old_window_size = self.get_size()
        self.isResizing = False

        # Default settings
        self.books_per_page = int(self._window.preferences.get_value("library_books_per_page"))
        self.library_layout = int(self._window.preferences.get_value("library_layout"))
        self.books_start_number = 1
        self.filter_title = self.filter_characters = self.filter_authors = ''
        self.filter_license = self.filter_genres = self.filter_languages = self.filter_series = self.filter_rating = self.filter_publisher = self.filter_publishdate = self.filter_read = self.filter_frames = self.filter_im_formats = 0
        self.read = ["Do Not Filter", "True", "False"]
        self.frames_def = ["Do Not Filter", "True", "False"]
        self.populate_filter_lists()
        self.custom_filters = gtk.combo_box_new_text()

        # action area
        self.toggle_button_hbox = gtk.HBox(True, 0)
        button = gtk.Button("N")
        button.set_tooltip_text('Normal Layout')
        if self.library_layout == 0:
          button.set_sensitive(False)
        button.connect("clicked", self.change_library_layout, "N")
        self.toggle_button_hbox.pack_start(button, True, True, 0)
        button = gtk.Button("C")
        button.set_tooltip_text('Compact Layout')
        if self.library_layout == 1:
          button.set_sensitive(False)
        button.connect("clicked", self.change_library_layout, "C")
        self.toggle_button_hbox.pack_start(button, True, True, 0)
        button = gtk.Button("L")
        button.set_tooltip_text('List Layout')
        if self.library_layout == 2:
          button.set_sensitive(False)
        button.connect("clicked", self.change_library_layout, "L")
        self.toggle_button_hbox.pack_start(button, True, True, 0)

        self.get_action_area().add(self.toggle_button_hbox)

        self.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
        self.get_action_area().set_layout(gtk.BUTTONBOX_EDGE)
        
        # Toolbar
        self.toolbar = gtk.Toolbar()
        self.toolbar.set_orientation(gtk.ORIENTATION_HORIZONTAL)
        self.toolbar.set_style(gtk.TOOLBAR_ICONS)
        if self._window.preferences.get_value("hidpi") == 'True':
          self.toolbar.set_icon_size(gtk.ICON_SIZE_DIALOG)
        else:
          self.toolbar.set_icon_size(gtk.ICON_SIZE_SMALL_TOOLBAR)
        self.toolbar.set_border_width(5)

        tool_button = gtk.ToolButton()
        tool_button.set_stock_id(gtk.STOCK_ADD)
        tool_button.set_tooltip_text('Add Comic Book')
        tool_button.connect("clicked", self.add_book)
        self.toolbar.insert(tool_button, 0)

        tool_button = gtk.ToolButton()
        tool_button.set_stock_id(gtk.STOCK_DIRECTORY)
        tool_button.set_tooltip_text('Add Comic Book Folder')
        tool_button.connect("clicked", self.add_book_folder)
        self.toolbar.insert(tool_button, 1)

        self.toolbar.insert(gtk.SeparatorToolItem(), 2)

        self.library_info_button = gtk.ToolButton()
        self.library_info_button.set_stock_id(gtk.STOCK_INFO)
        self.library_info_button.set_tooltip_text('Library Info')
        self.library_info_button.connect("clicked", self.show_info)
        self.toolbar.insert(self.library_info_button, 3)

        self.toolbar.insert(gtk.SeparatorToolItem(), 4)

        self.goto_first_button = gtk.ToolButton()
        self.goto_first_button.set_stock_id(gtk.STOCK_GOTO_FIRST)
        self.goto_first_button.connect("clicked", self.goto_first_page)
        self.toolbar.insert(self.goto_first_button, 5)

        self.goto_prev_button = gtk.ToolButton()
        self.goto_prev_button.set_stock_id(gtk.STOCK_GO_BACK)
        self.goto_prev_button.connect("clicked", self.goto_previous_page)
        self.toolbar.insert(self.goto_prev_button, 6)

        self.entry = gtk.Entry()
        self.entry.set_width_chars(7)
        self.entry.set_text('X/X')
        self.entry.set_sensitive(False)
        self.entry.show()
        entry_toolitem = gtk.ToolItem()
        entry_toolitem.add(self.entry)
        self.toolbar.insert(entry_toolitem, 7)

        self.goto_next_button = gtk.ToolButton()
        self.goto_next_button.set_stock_id(gtk.STOCK_GO_FORWARD)
        self.goto_next_button.connect("clicked", self.goto_next_page)
        self.toolbar.insert(self.goto_next_button, 8)

        self.goto_last_button = gtk.ToolButton()
        self.goto_last_button.set_stock_id(gtk.STOCK_GOTO_LAST)
        self.goto_last_button.connect("clicked", self.goto_last_page)
        self.toolbar.insert(self.goto_last_button, 9)

        self.toolbar.insert(gtk.SeparatorToolItem(), 10)

        tool_button = gtk.MenuToolButton(gtk.STOCK_PAGE_SETUP)
        tool_button.set_tooltip_text('Set Filter')
        tool_button.set_arrow_tooltip_text('List of Custom Filters')
        tool_button.connect("clicked", self.set_filter)

        self.filter_menu = gtk.Menu()
        self.load_custom_filters()
        tool_button.set_menu(self.filter_menu)
        self.toolbar.insert(tool_button, 11)

        sort_by_box = gtk.HBox(False, 0)
        sort_by_box.set_border_width(2)
        label = gtk.Label()
        label.set_markup('<b>Sort by:</b>')
        self.sort_by = gtk.combo_box_new_text()
        for sort_item in ['Title', 'Series', 'Author(s)', 'Publisher', 'Publish Date', 'Languages', 'Rating']:
          self.sort_by.append_text(sort_item)
        self.sort_by.set_active(int(self._window.preferences.get_value("library_default_sort_order")))
        self.sort_by.connect('changed', self.change_sort_order)
        sort_by_box.pack_start(label, False, False, 3)
        sort_by_box.pack_start(self.sort_by, False, False, 3)
        sort_toolitem = gtk.ToolItem()
        sort_toolitem.add(sort_by_box)
        self.toolbar.insert(sort_toolitem, 12)

        self.toolbar.show()
        self.vbox.pack_start(self.toolbar, False, False, 0)

        # book list window
        self.scrolled = gtk.ScrolledWindow()
        self.vbox.pack_start(self.scrolled, True, True, 0)

        self.book_list = gtk.VBox(False, 5)
        self.scrolled.add_with_viewport(self.book_list)

        # check if all books exist on their path
        if self._window.preferences.get_value("library_cleanup") == "True":
          self.check_books()

        self.total_books = len(self.library.tree.findall("book"))
        self.number_of_pages = int(round((self.total_books/float(self.books_per_page)) + 0.49999, 0))
        self.entry.set_width_chars(len(str(self.number_of_pages))*2 + 2)

        self.change_sort_order()
        self.scrolled.grab_focus()

        # show it
        self.show_all()
        
        self.run()
        self.destroy()
        
    def resize_library(self, *args):
        if self.library_layout == 1 and self.old_window_size[0] != self.get_size()[0] and not self.isResizing:
          self.isResizing = True
          self.old_window_size = self.get_size()
          self.display_books()
          self.show_all()
          self.isResizing = False

    def key_pressed(self, widget, event):
        if event.keyval == gtk.keysyms.Left and self.books_start_number > 1:
          self.goto_previous_page()
        elif event.keyval in (gtk.keysyms.Right, gtk.keysyms.space) and self.books_start_number < self.total_books - self.books_per_page + 1:
          self.goto_next_page()
        elif event.keyval == gtk.keysyms.Home:
          self.goto_first_page()
        elif event.keyval == gtk.keysyms.End:
          self.goto_last_page()

        self.scrolled.grab_focus()
        return

    def show_info(self, *args):
        library_dialog = library_info.LibraryInfoDialog(self)
        return

    def remove_custom_filter(self, widget):
        message = gtk.MessageDialog(parent=None, flags=0, type=gtk.MESSAGE_QUESTION, buttons=gtk.BUTTONS_YES_NO, message_format=None)
        message.set_markup("Are you sure you want to remove this filter from the list?\n\n" + '<b>' + self.custom_filters.get_active_text() + '</b>')
        response = message.run()
        if response == gtk.RESPONSE_YES:
          self._window.preferences.remove_library_filter(self.custom_filters.get_active_text())
        self._window.preferences.save_preferences()
        self.load_custom_filters()
        self.remove_button.set_sensitive(False)
        message.destroy()


    def save_custom_filter(self, widget):
        naming_dialog = gtk.Dialog('Save Filter as ...', self, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_OK, gtk.RESPONSE_OK))
        naming_dialog.set_resizable(False)
        naming_dialog.set_border_width(8)

        # filter name
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('New Filter Name: ')
        hbox.pack_start(label, False, False, 0)

        entry = gtk.Entry()
        entry.set_width_chars(30)
        if self.custom_filters.get_active_text() != None:
          entry.set_text(self.custom_filters.get_active_text())
        entry.show()
        hbox.pack_start(entry, False, False, 0)

        naming_dialog.vbox.pack_start(hbox, False, False, 0)
        naming_dialog.show_all()
        response = naming_dialog.run()

        # save filter to preferences
        if response == gtk.RESPONSE_OK and entry.get_text() != '':
          self._window.preferences.save_library_filter(entry.get_text(),
                                                       unicode(self.title_entry.get_text()),
                                                       unicode(self.authors_entry.get_text()),
                                                       unicode(self.series_filter.get_active_text()),
                                                       self.genres_filter.get_active_text(),
                                                       str(self.rating_filter.get_active()),
                                                       unicode(self.characters_entry.get_text()),
                                                       self.languages_filter.get_active_text(),
                                                       self.publishdate_filter.get_active_text(),
                                                       unicode(self.publisher_filter.get_active_text()),
                                                       unicode(self.license_filter.get_active_text()),
                                                       str(self.read_filter.get_active()),
                                                       str(self.frames_filter.get_active()),
                                                       str(self.im_formats_filter.get_active_text())
                                                       )
          self._window.preferences.save_preferences()
          self.load_custom_filters()
        naming_dialog.destroy()
        
        return

    def set_custom_filter(self, widget, title, authors, series, genres, rating, characters, languages, publishdate, publisher, license, read, has_frames, im_format):
        self.filter_title = title
        self.filter_authors = authors
        
        for idx, item in enumerate(self.series):
          if item == series:
            self.filter_series = idx

        if license != None:
          for idx, item in enumerate(self.licenses):
            if item == license:
              self.filter_license = idx
        else:
          self.filter_license = 0

        for idx, item in enumerate(self.genres):
          if item == genres:
            self.filter_genres = idx

        self.filter_rating = int(rating)

        if read != None:
          self.filter_read = int(read)
        else:
          self.filter_read = 0

        if has_frames != None:
          self.filter_frames = int(has_frames)
        else:
          self.filter_frames = 0

        self.filter_characters = characters

        for idx, item in enumerate(self.languages):
          if item == languages:
            self.filter_languages = idx

        for idx, item in enumerate(self.publishdates):
          if item == publishdate:
            self.filter_publishdate = idx

        for idx, item in enumerate(self.publishers):
          if item == publisher:
            self.filter_publisher = idx

        for idx, item in enumerate(self.im_formats):
          if item == im_format:
            self.filter_im_formats = idx
        
        if widget.get_name() == 'GtkMenuItem':
          self.goto_first_page()

    def set_custom_filter_combo(self, widget):
      self.remove_button.set_sensitive(True)
      for custom_filter in self._window.preferences.tree.find('library_custom_filters').findall('filter'):
        if custom_filter.get("name") == widget.get_active_text():
          self.set_custom_filter(widget, custom_filter.get("title"), custom_filter.get("authors"), custom_filter.get("series"),
                                 custom_filter.get("genres"), custom_filter.get("rating"), custom_filter.get("characters"),
                                 custom_filter.get("languages"), custom_filter.get("publishdate"), custom_filter.get("publisher"),
                                 custom_filter.get("license"), custom_filter.get("read"), custom_filter.get("has_frames"),
                                 custom_filter.get("im_format"))
          self.title_entry.set_text(self.filter_title)
          self.authors_entry.set_text(self.filter_authors)
          self.series_filter.set_active(self.filter_series)
          self.genres_filter.set_active(self.filter_genres)
          self.rating_filter.set_active(self.filter_rating)
          self.characters_entry.set_text(self.filter_characters)
          self.languages_filter.set_active(self.filter_languages)
          self.publishdate_filter.set_active(self.filter_publishdate)
          self.publisher_filter.set_active(self.filter_publisher)
          self.license_filter.set_active(self.filter_license)
          self.read_filter.set_active(self.filter_read)
          self.frames_filter.set_active(self.filter_frames)
          self.im_formats_filter.set_active(self.filter_im_formats)

    def load_custom_filters(self, *args):
        for filter_menu_item in self.filter_menu.get_children():
          filter_menu_item.destroy()

        for i in range(len(self._window.preferences.tree.find('library_custom_filters').findall('filter')) + 1):
          self.custom_filters.remove_text(0)

        for custom_filter in sorted(self._window.preferences.tree.find('library_custom_filters').findall('filter'), key=lambda order: order.get("name")):
          menuitem = gtk.MenuItem(custom_filter.get("name"))
          menuitem.connect("activate", self.set_custom_filter, custom_filter.get("title"), custom_filter.get("authors"), custom_filter.get("series"),
                           custom_filter.get("genres"), custom_filter.get("rating"), custom_filter.get("characters"), custom_filter.get("languages"),
                           custom_filter.get("publishdate"), custom_filter.get("publisher"), custom_filter.get("license"), custom_filter.get("read"),
                           custom_filter.get("has_frames"), custom_filter.get("im_format"))
          self.filter_menu.add(menuitem)
          self.custom_filters.append_text(custom_filter.get("name"))


        self.filter_menu.show_all()

    def populate_filter_lists(self, *args):
        self.languages = []
        self.genres = []
        self.series = []
        self.licenses = []
        self.publishers = []
        self.publishdates = []
        self.im_formats = []
        
        for book in self.library.tree.findall("book"):
          for language in get_element_text2(book, "languages").split(', '):
            if language not in self.languages and language != '??(no text layer)':
              self.languages.append(language)
          for genre in get_element_text2(book, "genres").split(', '):
            if genre not in self.genres and genre != '':
              self.genres.append(genre)
          for serie in get_element_text2(book, "sequence").split('), '):
            if unescape(serie[:serie.rfind('(') - 1]) not in self.series and serie != '':
              self.series.append(unescape(serie[:serie.rfind('(') - 1]))
          if get_element_text2(book, "publisher") not in self.publishers and get_element_text2(book, "publisher") != '':
              self.publishers.append(get_element_text2(book, "publisher"))
          if get_element_text2(book, "license") not in self.licenses and get_element_text2(book, "license") != '':
              self.licenses.append(get_element_text2(book, "license"))
          if get_element_text2(book, "publish_date")[0:4] not in self.publishdates and get_element_text2(book, "publish_date") != '':
              self.publishdates.append(get_element_text2(book, "publish_date")[0:4])
          if get_element_text2(book, "im_formats") not in self.im_formats and get_element_text2(book, "im_formats") != '':
              self.im_formats.append(get_element_text2(book, "im_formats"))

        self.languages.sort()
        self.languages.append('??(no text layer)')
        self.languages.insert(0, 'all languages')

        self.genres.sort()
        self.genres.insert(0, 'all genres')

        self.series.sort()
        self.series.insert(0, 'all series')

        self.licenses.sort()
        self.licenses.insert(0, 'all licenses')

        self.publishers.sort()
        self.publishers.insert(0, 'all publishers')

        self.im_formats.sort()
        self.im_formats.insert(0, 'all formats')

        self.publishdates.sort()
        self.publishdates.insert(0, 'all publish dates')

    def set_filter(self, *args):
        self.filter_dialog = gtk.Dialog('Set Filter', self, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                          (gtk.STOCK_APPLY, gtk.RESPONSE_OK))
        self.filter_dialog.set_resizable(False)
        self.filter_dialog.set_border_width(8)

        # clear filter button
        clear_button = gtk.Button(stock=gtk.STOCK_CLEAR)
        clear_button.connect("clicked", self.clear_filter)
        self.filter_dialog.get_action_area().add(clear_button)
        self.filter_dialog.get_action_area().reorder_child(clear_button, 0)

        self.filter_dialog.connect('key_press_event', self.filter_key_pressed)

        # populate lists
        self.populate_filter_lists()

        # customs filters
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        self.custom_filters = gtk.combo_box_new_text()
        self.custom_filters.set_tooltip_text('List of Custom Filters')
        self.load_custom_filters()
        self.custom_filters.connect("changed", self.set_custom_filter_combo)
        hbox.pack_start(self.custom_filters, False, False, 0)

        save_button = gtk.ToolButton(gtk.STOCK_SAVE)
        save_button.set_tooltip_text('Save Current Filter As ...')
        save_button.connect("clicked", self.save_custom_filter)
        hbox.pack_start(save_button, False, False, 0)

        self.remove_button = gtk.ToolButton(gtk.STOCK_REMOVE)
        self.remove_button.set_tooltip_text('Remove Custom Filter')
        self.remove_button.connect("clicked", self.remove_custom_filter)
        self.remove_button.set_sensitive(False)
        hbox.pack_start(self.remove_button, False, False, 0)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # separator
        separator = gtk.HSeparator()
        self.filter_dialog.vbox.pack_start(separator, False, False, 0)


        # Title
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Title (contains): ')
        hbox.pack_start(label, False, False, 0)

        self.title_entry = gtk.Entry()
        self.title_entry.set_width_chars(30)
        self.title_entry.set_text(self.filter_title)
        self.title_entry.show()
        hbox.pack_start(self.title_entry, False, False, 0)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # Authors
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Author(s) (contain): ')
        hbox.pack_start(label, False, False, 0)

        self.authors_entry = gtk.Entry()
        self.authors_entry.set_width_chars(30)
        self.authors_entry.set_text(self.filter_authors)
        self.authors_entry.show()
        hbox.pack_start(self.authors_entry, False, False, 0)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # Series
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Series: ')
        hbox.pack_start(label, False, False, 0)

        self.series_filter = gtk.combo_box_new_text()
        for serie in self.series:
          self.series_filter.append_text(serie)
        self.series_filter.set_active(self.filter_series)
        hbox.pack_start(self.series_filter, False, False, 0)
        self.series_filter.connect('changed', self.set_series_filter)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # Genres & Rating
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Genre: ')
        hbox.pack_start(label, False, False, 0)

        self.genres_filter = gtk.combo_box_new_text()
        for genre in self.genres:
          self.genres_filter.append_text(genre)
        self.genres_filter.set_active(self.filter_genres)
        hbox.pack_start(self.genres_filter, False, False, 0)
        self.genres_filter.connect('changed', self.set_genres_filter)

        label = gtk.Label()
        label.set_markup('    Rating: ')
        hbox.pack_start(label, False, False, 0)

        self.rating_filter = gtk.combo_box_new_text()
        self.rating_filter.append_text('Do not filter')
        self.rating_filter.append_text('Without rating')
        self.rating_filter.append_text('*')
        self.rating_filter.append_text('**')
        self.rating_filter.append_text('***')
        self.rating_filter.append_text('****')
        self.rating_filter.append_text('*****')
        self.rating_filter.set_active(self.filter_rating)
        hbox.pack_start(self.rating_filter, False, False, 0)
        self.rating_filter.connect('changed', self.set_rating_filter)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # Characters & read
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Characters (contain): ')
        hbox.pack_start(label, False, False, 0)

        self.characters_entry = gtk.Entry()
        self.characters_entry.set_width_chars(15)
        self.characters_entry.set_text(self.filter_characters)
        self.characters_entry.show()
        hbox.pack_start(self.characters_entry, False, False, 0)

        label = gtk.Label()
        label.set_markup('    Read: ')
        hbox.pack_start(label, False, False, 0)

        self.read_filter = gtk.combo_box_new_text()
        for read in self.read:
          self.read_filter.append_text(read)
        self.read_filter.set_active(self.filter_read)
        hbox.pack_start(self.read_filter, False, False, 0)
        self.read_filter.connect('changed', self.set_read_filter)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # Languages & Frames
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Language: ')
        hbox.pack_start(label, False, False, 0)

        self.languages_filter = gtk.combo_box_new_text()
        for language in self.languages:
          self.languages_filter.append_text(language)
        self.languages_filter.set_active(self.filter_languages)
        hbox.pack_start(self.languages_filter, False, False, 0)
        self.languages_filter.connect('changed', self.set_languages_filter)

        label = gtk.Label()
        label.set_markup('    Frames definition: ')
        hbox.pack_start(label, False, False, 0)

        self.frames_filter = gtk.combo_box_new_text()
        for frame_def in self.frames_def:
          self.frames_filter.append_text(frame_def)
        self.frames_filter.set_active(self.filter_frames)
        hbox.pack_start(self.frames_filter, False, False, 0)
        self.frames_filter.connect('changed', self.set_frames_filter)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # Publisher & Publish date
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Publisher: ')
        hbox.pack_start(label, False, False, 0)

        self.publisher_filter = gtk.combo_box_new_text()
        for publisher in self.publishers:
          self.publisher_filter.append_text(publisher)
        self.publisher_filter.set_active(self.filter_publisher)
        hbox.pack_start(self.publisher_filter, False, False, 0)
        self.publisher_filter.connect('changed', self.set_publisher_filter)

        label = gtk.Label()
        label.set_markup('    Publish date: ')
        hbox.pack_start(label, False, False, 0)

        self.publishdate_filter = gtk.combo_box_new_text()
        for publishdate in self.publishdates:
          self.publishdate_filter.append_text(publishdate)
        self.publishdate_filter.set_active(self.filter_publishdate)
        hbox.pack_start(self.publishdate_filter, False, False, 0)
        self.publishdate_filter.connect('changed', self.set_publishdate_filter)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # License
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('License: ')
        hbox.pack_start(label, False, False, 0)

        self.license_filter = gtk.combo_box_new_text()
        for license in self.licenses:
          self.license_filter.append_text(license)
        self.license_filter.set_active(self.filter_license)
        hbox.pack_start(self.license_filter, False, False, 0)
        self.license_filter.connect('changed', self.set_license_filter)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)

        # Image Formats
        hbox = gtk.HBox(False, 0)
        hbox.set_border_width(5)

        label = gtk.Label()
        label.set_markup('Image Format: ')
        hbox.pack_start(label, False, False, 0)

        self.im_formats_filter = gtk.combo_box_new_text()
        for im_format in self.im_formats:
          self.im_formats_filter.append_text(im_format)
        self.im_formats_filter.set_active(self.filter_im_formats)
        hbox.pack_start(self.im_formats_filter, False, False, 0)
        self.im_formats_filter.connect('changed', self.set_im_formats_filter)

        self.filter_dialog.vbox.pack_start(hbox, False, False, 0)
        
        # show it
        self.filter_dialog.show_all()
        response = self.filter_dialog.run()

        if response == gtk.RESPONSE_OK:
          self.filter_title = self.title_entry.get_text()
          self.filter_characters = self.characters_entry.get_text()
          self.filter_authors = self.authors_entry.get_text()
        
        self.filter_dialog.destroy()

        if response == gtk.RESPONSE_OK:
          self.goto_first_page()
        
        return

    def filter_key_pressed(self, widget, event):
        if event.keyval in (gtk.keysyms.Return, gtk.keysyms.KP_Enter):
          self.filter_dialog.response(gtk.RESPONSE_OK)

    def set_series_filter(self, widget, *args):
        self.filter_series = widget.get_active()
        return

    def set_license_filter(self, widget, *args):
        self.filter_license = widget.get_active()
        return

    def set_publishdate_filter(self, widget, *args):
        self.filter_publishdate = widget.get_active()
        return

    def set_publisher_filter(self, widget, *args):
        self.filter_publisher = widget.get_active()
        return

    def set_rating_filter(self, widget, *args):
        self.filter_rating = widget.get_active()
        return

    def set_read_filter(self, widget, *args):
        self.filter_read = widget.get_active()
        return

    def set_frames_filter(self, widget, *args):
        self.filter_frames = widget.get_active()
        return

    def set_genres_filter(self, widget, *args):
        self.filter_genres = widget.get_active()
        return

    def set_languages_filter(self, widget, *args):
        self.filter_languages = widget.get_active()
        return

    def set_im_formats_filter(self, widget, *args):
        self.filter_im_formats = widget.get_active()
        return

    def clear_filter(self, widget):
        self.title_entry.set_text('')
        self.authors_entry.set_text('')
        self.characters_entry.set_text('')
        self.series_filter.set_active(0)
        self.license_filter.set_active(0)
        self.genres_filter.set_active(0)
        self.rating_filter.set_active(0)
        self.read_filter.set_active(0)
        self.frames_filter.set_active(0)
        self.languages_filter.set_active(0)
        self.publishdate_filter.set_active(0)
        self.publisher_filter.set_active(0)
        self.im_formats_filter.set_active(0)
        self.filter_title = self.filter_characters = self.filter_authors = ''
        return

    def change_library_layout(self, widget, layout):
        if layout == 'N':
          self.library_layout = 0
        elif layout == 'C':
          self.library_layout = 1
        elif layout == 'L':
          self.library_layout = 2
        for button in self.toggle_button_hbox.get_children():
          if button.get_label() != layout:
            button.set_sensitive(True)
          else:
            button.set_sensitive(False)
        self.display_books()
        self.update_entry()
        self.show_all()

    def change_sort_order(self, *args):
        data = []
        sort_items = ['title', 'sequence', 'authors', 'publisher', 'publish_date', 'languages', 'rating']
        for elem in self.library.tree:
          key = elem.findtext(sort_items[self.sort_by.get_active()])
          if key == None:
            key = '~'
          else:
            if key == '' or key == '??(no text layer)': # put it at the end of the sorted list
              key = '~'
            elif sort_items[self.sort_by.get_active()] == 'sequence' and key != '|': # sort by serie name as string than by serie number as number
              key = key[:key.rfind('(') - 1] + key[key.rfind('(') + 1:key.rfind(')')].rjust(4, '0')
            elif sort_items[self.sort_by.get_active()] == 'rating': # sort descending
              key = str(5 - int(elem.findtext(sort_items[self.sort_by.get_active()])))

          data.append((key, elem))
        data.sort()

        self.library.tree[:] = [item[-1] for item in data]
        self.display_books()
        self.update_entry()
        if self.total_books > 0:
          self.library_info_button.set_sensitive(True)
        else:
          self.library_info_button.set_sensitive(False)
        self.show_all()

        return

    def update_entry(self, *args):
        current_page = str(int(round((self.books_start_number - 1)/self.books_per_page + 1, 0)))
        entry_text = ' ' + current_page + '/' + str(self.number_of_pages) + ' '
        self.entry.set_text(entry_text)

    def goto_first_page(self, *args):
        self.books_start_number = 1
        self.display_books()
        self.update_entry()
        self.show_all()
        return

    def goto_previous_page(self, *args):
        self.books_start_number = self.books_start_number - self.books_per_page
        self.update_entry()
        self.display_books()
        self.show_all()
        return

    def goto_next_page(self, *args):
        self.books_start_number = self.books_start_number + self.books_per_page
        self.update_entry()
        self.display_books()
        self.show_all()
        return

    def goto_last_page(self, *args):
        self.books_start_number = (self.number_of_pages - 1) * self.books_per_page + 1
        self.update_entry()
        self.display_books()
        self.show_all()
        return

    def open_book(self, widget, event, filename):
        dialog = gtk.Dialog('Comic Book', self, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                          (gtk.STOCK_REMOVE, gtk.RESPONSE_REJECT, gtk.STOCK_OK, gtk.RESPONSE_OK))
        dialog.set_resizable(False)
        dialog.set_size_request(700 * self._window.ui_scale_factor, -1)
        dialog.set_border_width(8)

        hbox = gtk.HBox(False, 0)

        open_button = gtk.Button(stock=gtk.STOCK_OPEN)
        dialog.add_action_widget(open_button, gtk.RESPONSE_APPLY)
        open_button.grab_focus()

        for book in self.library.tree.findall("book"):
          if book.get("path") == filename:
            # coverpage
            coverpage = gtk.Image()
            cover_pix = gtk.gdk.pixbuf_new_from_file_at_size(book.find("coverpage").text, 200 * self._window.ui_scale_factor, 200 * self._window.ui_scale_factor)
            coverpage.set_from_pixbuf(cover_pix)
            coverpage.set_alignment(0, 0)

            vbox = gtk.VBox(False, 5)
            vbox.set_border_width(8)
            vbox.pack_start(coverpage, False, False, 0)

            # rating
            self.rating_box = gtk.HBox(False, 5)
            self.rating_box.rating = int(get_element_text2(book, "rating"))

            self.rate_book(None, None, filename, self.rating_box.rating)

            vbox.pack_start(self.rating_box, False, False, 0)

            # progress
            x_progress = 0
            (page_number, frame_number, zoom_level, language_layer) = self._window.history.get_book_details(filename)
            if book.find("pages") != None and page_number > 1:
              progress = int(float(page_number)/(int(book.find("pages").text) + 1)*100)
            else:
              progress = 0

            if get_element_text2(book, "read") == "True":
              x_progress = 100
            else:
              x_progress = progress

            self.read_button = gtk.CheckButton('Read: ' + str(x_progress) + '%')
            self.read_button.connect("toggled", self.set_read_button, filename, progress)
            if get_element_text2(book, "read") == "True":
              self.read_button.set_active(True)

            vbox.pack_start(self.read_button, False, False, 0)

            hbox.pack_start(vbox, False, False, 0)


            scrolled = gtk.ScrolledWindow()
            scrolled.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
            scrolled.set_size_request(480, 300)
            book_details = gtk.VBox(False, 0)
            book_details.set_border_width(5)
            scrolled.add_with_viewport(book_details)

            # book meta-data
            label = gtk.Label()
              
            publish_date = " (" + get_element_text2(book, "publish_date")[0:4] + ")"

            bookname = get_element_text3(book, "title", "en")
            for title in book.findall("title"):
              if self.preferences.get_value("default_language") != "0":
                if title.get("lang") == constants.LANGUAGES[int(self.preferences.get_value("default_language"))]:
                  bookname = title.text

            annotation = get_element_text3(book, "annotation", "en")
            for anno in book.findall("annotation"):
              if self.preferences.get_value("default_language") != "0":
                if anno.get("lang") == constants.LANGUAGES[int(self.preferences.get_value("default_language"))]:
                  annotation = anno.text

            markup = '<big><b>' + bookname + publish_date + '</b></big>' + '\n<b>Author(s): </b> ' + get_element_text2(book, "authors")
            if get_element_text2(book, "sequence") != '':
              markup = markup + '\n<b>Series: </b> ' + get_element_text2(book, "sequence")
            markup = markup + '\n<b>Publisher: </b> ' + get_element_text2(book, "publisher") 
            if get_element_text2(book, "license") != '':
              markup = markup + '\n<b>License: </b> ' + get_element_text2(book, "license")
            markup = markup + '\n<b>Genre(s): </b> ' + get_element_text2(book, "genres")
            if get_element_text2(book, "characters") != '':
              markup = markup + '\n<b>Characters: </b> ' + get_element_text2(book, "characters")
            markup = markup + '\n<b>Annotation: </b> ' + annotation + '\n<b>Frames definitions: </b> '
            if get_element_text2(book, "has_frames") == '':
              markup = markup + 'False'
            else:
              markup = markup + get_element_text2(book, "has_frames")
            markup = markup + '\n<b>Languages: </b> ' + get_element_text2(book, "languages")
            if get_element_text2(book, "im_formats") != '':
              markup = markup + '\n<b>Image Format(s): </b> ' + get_element_text2(book, "im_formats")

            label.set_markup(markup)
            book_details.pack_start(label, False)

            for label in book_details.get_children():
              label.set_alignment(0, 0)
              label.set_line_wrap(True)
              label.set_selectable(True)

            hbox.pack_end(scrolled, True, True, 0)
            break

        dialog.vbox.pack_start(hbox, False, False, 0)
        dialog.show_all()
        response = dialog.run()

        if response == gtk.RESPONSE_APPLY:
          self._window.filename = filename
          self._window.original_filename = filename
          dialog.destroy()
          self.destroy()
        elif response == gtk.RESPONSE_REJECT:
          self.remove_book(filename, bookname)
          dialog.destroy()
        elif response == gtk.RESPONSE_OK:
          dialog.destroy()
          if self.rating_box.rating != int(get_element_text2(book, "rating")):
            self.library.set_value("rating", str(self.rating_box.rating), filename)
            self.change_sort_order()
          if ((get_element_text2(book, "read") != "True" and self.read_button.get_active()) or
              (get_element_text2(book, "read") == "True" and not self.read_button.get_active())):
            if self.read_button.get_active():
              self.library.set_value("read", "True", filename)
            else:
              self.library.set_value("read", "False", filename)
            self.change_sort_order()
        else:
          dialog.destroy()

        return

    def remove_book(self, filename, bookname):
        message = gtk.MessageDialog(parent=None, flags=0, type=gtk.MESSAGE_QUESTION, buttons=gtk.BUTTONS_YES_NO, message_format=None)
        message.set_markup("Are you sure you want to remove this book from library?\n\n" + '<b>' + bookname + '</b>')
        response = message.run()
        if response == gtk.RESPONSE_YES:
          self.library.delete_book(filename)
          self.library.save_library()
          self.display_books()
          self.update_entry()
          self.show_all()
        message.destroy()

    def set_read_button(self, widget, filename, progress):
        if widget.get_active():
          x_progress = 100
        else:
          x_progress = progress
        widget.set_label('Read: ' + str(x_progress) + '%')

    def rate_book(self, widget, event, filename, rating):
        self.rating_box.rating = rating

        for i in self.rating_box.get_children():
          i.destroy()

        for asterisk in range(rating):
          star = gtk.Image()
          if self._window.preferences.get_value("hidpi") == 'True':
            star.set_from_stock(gtk.STOCK_ABOUT, gtk.ICON_SIZE_DND)
          else:
            star.set_from_stock(gtk.STOCK_ABOUT, gtk.ICON_SIZE_MENU)
          star.set_tooltip_text('Rating')
          eventbox = gtk.EventBox()
          eventbox.add(star)
          eventbox.connect('button-press-event', self.rate_book, filename, asterisk + 1)
          self.rating_box.pack_start(eventbox, False, False, 0)

        for asterisk in range(5 - rating):
          blank = gtk.Image()
          if self._window.preferences.get_value("hidpi") == 'True':
            blank.set_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_DND)
          else:
            blank.set_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
          blank.set_tooltip_text('Rating')
          eventbox = gtk.EventBox()
          eventbox.add(blank)
          eventbox.connect('button-press-event', self.rate_book, filename, asterisk + rating + 1)
          self.rating_box.pack_start(eventbox, False, False, 0)

        self.rating_box.show_all()

    def check_books(self, *args):
        deleted = False
        for book in self.library.tree.findall("book"):
          if not os.path.exists(book.get("path")):
            self.library.delete_book(unicode(book.get("path")))
            deleted = True

        # remove duplicates (cbz with same base filename as acbf file)
        referenced_archives = []
        for referenced_archive in self.library.tree.findall("book"):
          if referenced_archive.get("path")[-4:].upper() == 'ACBF':
            referenced_archives.append(referenced_archive.get("path")[0:-4] + 'cbz')

        for referenced_archive in referenced_archives:
          book_is_added = True
          self.library.delete_book(referenced_archive)

        if deleted:
          self.library.save_library()

        # cleanup library covers directory
        if randint(0,49) == 0: #this is slow, therefore we do it once in 50 times
          for root, dirs, files in os.walk(os.path.join(constants.CONFIG_DIR, 'Covers')):
            for f in files:
              do_remove = True
              for book in self.library.tree.findall("book"):
                if os.path.join(root, f) == book.find("coverpage").text:
                  do_remove = False
              if do_remove:
                os.unlink(os.path.join(root, f))
            for d in dirs:
              if not os.listdir(os.path.join(root, d)):
                shutil.rmtree(os.path.join(root, d))

        return

    def display_books(self, *args):
        for child_widget in self.book_list.get_children():
          child_widget.destroy()
        while gtk.events_pending():
          gtk.main_iteration()

        horizontal_list = gtk.HBox(False, 0)
        current_line_length = 0

        books_added = 0
        self.total_books = 0

        if self.library_layout == 2:
          self.scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        else:
          self.scrolled.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)

        for book in self.library.tree.findall("book"):
          book_series = get_element_text2(book, "sequence").split('), ')

          for i in range(len(book_series)):
            book_series[i] = unescape(book_series[i][:book_series[i].rfind('(') - 1])

          if get_element_text2(book, "read") == '':
            read = "False"
          else:
            read = get_element_text2(book, "read")

          if get_element_text2(book, "has_frames") == '':
            has_frames = "False"
          else:
            has_frames = get_element_text2(book, "has_frames")

          if ((self.filter_genres == 0 or self.genres[self.filter_genres] in get_element_text2(book, "genres").split(', ')) and
              (self.filter_languages == 0 or self.languages[self.filter_languages] in get_element_text2(book, "languages").split(', ')) and
              (self.filter_title == '' or self.filter_title.upper() in get_element_text2(book, "title").upper()) and
              (self.filter_characters == '' or self.filter_characters.upper() in get_element_text2(book, "characters").upper()) and
              (self.filter_authors == '' or self.filter_authors.upper() in get_element_text2(book, "authors").upper()) and
              (self.filter_series == 0 or self.series[self.filter_series] in book_series) and
              (self.filter_rating == 0 or str(self.filter_rating - 1) == get_element_text2(book, "rating")) and
              (self.filter_read == 0 or self.read[self.filter_read] == read) and
              (self.filter_frames == 0 or self.frames_def[self.filter_frames] == has_frames) and
              (self.filter_im_formats == 0 or self.im_formats[self.filter_im_formats] in get_element_text2(book, "im_formats")) and
              (self.filter_publisher == 0 or self.publishers[self.filter_publisher] == get_element_text2(book, "publisher")) and
              (self.filter_license == 0 or self.licenses[self.filter_license] == get_element_text2(book, "license")) and
              (self.filter_publishdate == 0 or self.publishdates[self.filter_publishdate] == get_element_text2(book, "publish_date")[0:4])
              ):
            self.total_books = self.total_books + 1

            if books_added < self.books_per_page and self.total_books >= self.books_start_number:
              hbox = gtk.HBox(False, 0)
              eventbox = gtk.EventBox()
              eventbox.connect('button-press-event', self.open_book, book.get("path"))

              if self.library_layout in (0, 1):
                # coverpage
                if book.find("coverpage").get("type") == 'link':
                  coverpage = gtk.Image()
                  cover_pix = gtk.gdk.pixbuf_new_from_file_at_size(book.find("coverpage").text, 200 * self._window.ui_scale_factor, 200 * self._window.ui_scale_factor)
                  coverpage_width = cover_pix.get_width()
                  coverpage.set_from_pixbuf(cover_pix)
                else:
                  decoded = base64.b64decode(book.find("coverpage").text)
                  coverpage_i = Image.open(cStringIO.StringIO(decoded))
                  coverpage_width = coverpage_i.size[0]
                  coverpage = gtk.Image()
                  cover_pix = pil_to_pixbuf(coverpage_i, '#000')
                  cover_pix = cover_pix.scale_simple(200 * self._window.ui_scale_factor, 200 * self._window.ui_scale_factor, gtk.gdk.INTERP_BILINEAR)
                  coverpage.set_from_pixbuf(cover_pix)
                coverpage.set_alignment(0, 0)
                eventbox.add(coverpage)

                vbox = gtk.VBox(False, 5)
                vbox.set_border_width(8)
                vbox.pack_start(eventbox, False, False, 0)

              if self.library_layout == 0:  
                hbox.pack_start(vbox, False, False, 0)

                book_details = gtk.VBox(False, 0)
                book_details.set_border_width(5)

                # book meta-data
                label = gtk.Label()
              
                publish_date = " (" + get_element_text2(book, "publish_date")[0:4] + ")"

                rating = '  '
                for asterisk in range(int(get_element_text2(book, "rating"))):
                  rating = rating + '*'

                bookname = get_element_text3(book, "title", "en")
                for title in book.findall("title"):
                  if self.preferences.get_value("default_language") != "0":
                    if title.get("lang") == constants.LANGUAGES[int(self.preferences.get_value("default_language"))]:
                      bookname = title.text

                annotation = get_element_text3(book, "annotation", "en")
                for anno in book.findall("annotation"):
                  if self.preferences.get_value("default_language") != "0":
                    if anno.get("lang") == constants.LANGUAGES[int(self.preferences.get_value("default_language"))]:
                      annotation = anno.text

                markup = '<big><b>' + bookname + publish_date + rating +  '</b></big>' + '\n<b>Author(s): </b> ' + get_element_text2(book, "authors")
                if get_element_text2(book, "sequence") != '':
                  markup = markup + '\n<b>Series: </b> ' + get_element_text2(book, "sequence")
                markup = markup + '\n<b>Publisher: </b> ' + get_element_text2(book, "publisher") 
                if get_element_text2(book, "license") != '':
                  markup = markup + '\n<b>License: </b> ' + get_element_text2(book, "license")
                markup = markup + '\n<b>Genre(s): </b> ' + get_element_text2(book, "genres")
                if get_element_text2(book, "characters") != '':
                  markup = markup + '\n<b>Characters: </b> ' + get_element_text2(book, "characters")
                markup = markup + '\n<b>Annotation: </b> ' + annotation + '\n<b>Frames definitions: </b> '
                if get_element_text2(book, "has_frames") == '':
                  markup = markup + 'False'
                else:
                  markup = markup + get_element_text2(book, "has_frames")
                markup = markup + '\n<b>Languages: </b> ' + get_element_text2(book, "languages")
                if get_element_text2(book, "im_formats") != '':
                  markup = markup + '\n<b>Image Format(s): </b> ' + get_element_text2(book, "im_formats")

                label.set_markup(markup)
                label.connect("size-allocate", self.size_request)
                book_details.pack_start(label, False)

                for label in book_details.get_children():
                  label.set_alignment(0, 0)
                  label.set_line_wrap(True)
                  label.set_selectable(True)

                hbox.pack_end(book_details, True, True, 0)
                self.book_list.pack_start(hbox, False, False, 0)
                books_added = books_added + 1
              elif self.library_layout == 1:
                if ((current_line_length + coverpage_width) > self.get_size()[0]):
                  current_line_length = coverpage_width + 30
                  self.book_list.pack_start(horizontal_list, False, False, 0)
                  horizontal_list = gtk.HBox(False, 0)
                else:
                  current_line_length = current_line_length + coverpage_width + 30

                label = gtk.Label()
                label.set_markup('<small><b>' + get_element_text2(book, "title") + '</b></small>')

                if len(get_element_text2(book, "title"))/float(coverpage_width) > 0.13: # small hack to make alignment work properly atleast on non-wrapped labels
                  label.set_line_wrap(True)
                label.set_justify(gtk.JUSTIFY_CENTER)
                label.set_size_request(coverpage_width, -1)
                label.set_alignment(0.5, 0.5)

                vbox.pack_start(label, False, False, 0)

                publish_date = " (" + get_element_text2(book, "publish_date")[0:4] + ")"

                series = get_element_text2(book, "sequence")
                if series != '':
                   series = '\nSeries: ' + series

                authors = get_element_text2(book, "authors")
                if authors != '':
                   authors = '\nAuthors: ' + authors

                rating = '  '
                for asterisk in range(int(get_element_text2(book, "rating"))):
                  rating = rating + '*'

                eventbox.set_tooltip_text(unescape(get_element_text2(book, "title") + publish_date + rating + series + authors))
                horizontal_list.pack_start(vbox, True, False, 0)
                books_added = books_added + 1
              elif self.library_layout == 2: #line layout
                eventbox.add(hbox)

                book_title = get_element_text2(book, "title")
                publish_date = get_element_text2(book, "publish_date")[0:4]
                rating = ''
                for asterisk in range(int(get_element_text2(book, "rating"))):
                  rating = rating + '*'
                authors = get_element_text2(book, "authors").split(',')[0].split('(')[0]
                series = get_element_text2(book, "sequence")
                publisher = get_element_text2(book, "publisher")
                genres = get_element_text2(book, "genres").split(',')[0]
                license = get_element_text2(book, "license")
                languages = get_element_text2(book, "languages")

                columns_list = [book_title, publish_date, rating, publisher, authors, series, genres, languages]

                if books_added == 0:
                  labels_hbox = gtk.HBox(False, 0)
                  labels_list = ["Title", "Publish Date", "Rating", "Publisher", "Authors", "Series", "Genres", "Languages"]
                  for column_label in labels_list:
                    label = gtk.Label()
                    label.set_markup(column_label)
                    label.set_alignment(0, 0.5)
                    labels_hbox.pack_start(label, False, False, 5)
                  self.book_list.pack_start(labels_hbox, False, False, 0)
                
                for column_label in columns_list:
                  label = gtk.Label()
                  label.set_markup(column_label)
                  label.set_alignment(0, 0.5)
                  hbox.pack_start(label, False, False, 5)

                if books_added % 2 == 0:
                  eventbox.set_state(gtk.STATE_ACTIVE)
                self.book_list.pack_start(eventbox, False, False, 0)
                books_added = books_added + 1

        if self.library_layout == 1:
          self.book_list.pack_start(horizontal_list, False, False, 0)
        elif self.library_layout == 2:
          # get max values
          book_column_lengths = [0, 0, 0, 0, 0, 0, 0, 0]
          for book in self.book_list.get_children():
            for i in range(0, len(book_column_lengths)):
              try:
                label = unicode(book.get_children()[0].get_children()[i].get_text())
              except:
                label = unicode(book.get_children()[i].get_text())
              if len(label) > book_column_lengths[i]:
                book_column_lengths[i] = len(label)

          # align columns
          for idx, book in enumerate(self.book_list.get_children()):
            for i in range(0, len(book_column_lengths)):
              try:
                label = unicode(book.get_children()[0].get_children()[i].get_text())
              except:
                label = unicode(book.get_children()[i].get_text())
              new_label = label.ljust(book_column_lengths[i])
              if idx == 0:
                new_label = "<span font_family='monospace' underline='low' font_weight='bold'>" + escape(new_label) + "</span>"
                book.get_children()[i].set_markup(new_label)
              else:
                book.get_children()[0].get_children()[i].set_markup("<tt>" + escape(new_label) + "</tt>")
            
        self.number_of_pages = int(round((self.total_books/float(self.books_per_page)) + 0.49999, 0))
        
        # adjust navigational buttons sensitivity
        if self.books_start_number == 1:
          self.goto_first_button.set_sensitive(False)
          self.goto_prev_button.set_sensitive(False)
        else:
          self.goto_first_button.set_sensitive(True)
          self.goto_prev_button.set_sensitive(True)

        if self.books_start_number < self.total_books - self.books_per_page + 1:
          self.goto_next_button.set_sensitive(True)
          self.goto_last_button.set_sensitive(True)
        else:
          self.goto_next_button.set_sensitive(False)
          self.goto_last_button.set_sensitive(False)

    def size_request(self, l, s ):
        l.set_size_request(s.width -1, -1)

    def add_book_folder(self, *args):
        filechooser = gtk.FileChooserDialog(title='Add Comic Book Folder ...', action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                  buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        filechooser.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)

        response = filechooser.run()
        if response != gtk.RESPONSE_OK:
          filechooser.destroy()
          return
        
        directory = unicode(filechooser.get_filename())
        filechooser.destroy()

        progress_dialog = gtk.Dialog('Adding Comic Books ...', self, gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT, None)
        progress_dialog.set_resizable(False)
        progress_dialog.set_border_width(8)
        progress_dialog.set_geometry_hints(min_height=100 * self._window.ui_scale_factor, min_width=400 * self._window.ui_scale_factor)

        progress_bar = gtk.ProgressBar()
        progress_bar.set_size_request(-1,13 * self._window.ui_scale_factor)
        progress_bar.show()

        progress_dialog.vbox.pack_start(progress_bar, False, False, 5)

        book_name = gtk.Label()
        book_name.set_markup('')
        progress_dialog.vbox.pack_start(book_name)

        progress_dialog.show_all()

        while gtk.events_pending():
          gtk.main_iteration()

        books_added = float(0)
        total_books_to_add = 0

        for root, dirs, files in os.walk(directory):
          for f in files:
            if f[-4:].upper() == '.CBZ' or f[-5:].upper() == '.ACBF' or f[-4:].upper() == '.ACV':
              total_books_to_add = total_books_to_add + 1

        for root, dirs, files in os.walk(directory):
          for f in files:
            f = unicode(f)
            if f[-4:].upper() == '.CBZ' or f[-5:].upper() == '.ACBF' or f[-4:].upper() == '.ACV':
              book_name.set_markup(escape(f))
              try:
                self.insert_new_book(os.path.join(root, f), False)
              except Exception as inst:
                message = gtk.MessageDialog(parent=None, flags=0, type=gtk.MESSAGE_WARNING, buttons=gtk.BUTTONS_OK, message_format=None)
                message.set_markup("Failed to import comic book\n\n" + '<b>' + f + '</b>\n\n' + 'Exception: %s' % inst)
                response = message.run()
                message.destroy()
              books_added = books_added + 1
              progress_bar.set_fraction(books_added/total_books_to_add)
              while gtk.events_pending():
                gtk.main_iteration()

        progress_dialog.destroy()

        self.library.save_library()
        self.populate_filter_lists()
        self.change_sort_order()
        self.show_all()
        
        return

    def add_book(self, *args):
        filechooser = gtk.FileChooserDialog(title='Add Comic Book ...', action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                  buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        # filters
        filter = gtk.FileFilter()
        filter.set_name("Comicbook files")
        filter.add_pattern("*.acbf")
        filter.add_pattern("*.acv")
        filter.add_pattern("*.cbz")
        filter.add_pattern("*.zip")
        filter.add_pattern("*.cbr")
        filechooser.add_filter(filter)

        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        filechooser.add_filter(filter)

        response = filechooser.run()

        if response != gtk.RESPONSE_OK:
          filechooser.destroy()
          return
        
        filename = unicode(filechooser.get_filename())
        filechooser.destroy()

        try:
          if self.insert_new_book(filename, True):
            self.library.save_library()
            self.populate_filter_lists()
            self.change_sort_order()
            self.show_all()
        except Exception as inst:
          message = gtk.MessageDialog(parent=None, flags=0, type=gtk.MESSAGE_WARNING, buttons=gtk.BUTTONS_OK, message_format=None)
          message.set_markup("Failed to import comic book\n\n" + '<b>' + filename + '</b>\n\n' + 'Exception: %s' % inst)
          response = message.run()
          message.destroy()

        return

    def insert_new_book(self, filename, show_dialog):
        # check if already exists in library
        rating = '0'
        for book in self.library.tree.findall("book"):
          if book.get("path") == filename:
            rating = self.library.get_value("rating", filename)
            self.library.delete_book(filename)
        
        coverpage, book_title, publish_date, publisher, authors, genres, sequence, annotation, languages, characters, pages, license, has_frames, im_formats = self.load_file(filename, show_dialog)

        if book_title == {}:
          return False

        new_book = xml.SubElement(self.library.tree, "book", path=filename)

        for title in book_title.items():
          new_book_title = xml.SubElement(new_book, "title", lang=title[0])
          new_book_title.text = title[1]

        new_authors = xml.SubElement(new_book, "authors")
        new_authors.text = authors

        new_publish_date = xml.SubElement(new_book, "publish_date")
        new_publish_date.text = publish_date

        new_publisher = xml.SubElement(new_book, "publisher")
        new_publisher.text = publisher

        new_sequence = xml.SubElement(new_book, "sequence")
        new_sequence.text = sequence

        for anno in annotation.items():
          new_annotation = xml.SubElement(new_book, "annotation", lang=anno[0])
          new_annotation.text = anno[1]

        new_languages = xml.SubElement(new_book, "languages")
        new_languages.text = languages

        new_genres = xml.SubElement(new_book, "genres")
        new_genres.text = genres

        new_characters = xml.SubElement(new_book, "characters")
        new_characters.text = characters

        new_rating = xml.SubElement(new_book, "rating")
        new_rating.text = rating

        new_pages = xml.SubElement(new_book, "pages")
        new_pages.text = str(pages)

        new_license = xml.SubElement(new_book, "license")
        new_license.text = license

        new_coverpage = xml.SubElement(new_book, "coverpage", type='link')
        new_coverpage.text = coverpage

        new_read_flag = xml.SubElement(new_book, "read")
        new_read_flag.text = "False"

        new_has_frames = xml.SubElement(new_book, "has_frames")
        new_has_frames.text = str(has_frames)

        new_im_formats = xml.SubElement(new_book, "im_formats")
        new_im_formats.text = str(im_formats)

        return True

    def load_file(self, in_filename, show_dialog):
        prepared_file = fileprepare.FilePrepare(self, in_filename, self._window.library_dir, show_dialog)
        filename = prepared_file.filename
        self.tempdir = self._window.library_dir
        acbf_document = acbfdocument.ACBFDocument(self, filename)

        if not acbf_document.valid:
          return None, None, None, None, None, None, None, None, None, None, None, None

        # coverpage
        coverpage = acbf_document.coverpage
        coverpage.thumbnail((int(coverpage.size[0]*150 * self._window.ui_scale_factor/float(coverpage.size[1])),150 * self._window.ui_scale_factor), Image.NEAREST)
        output_directory = os.path.join(os.path.join(constants.CONFIG_DIR, 'Covers'), acbf_document.book_title[acbf_document.book_title.items()[0][0]][0].upper())
        if not os.path.exists(output_directory):
          os.makedirs(output_directory, 0700)

        cover_number = 0
        for root, dirs, files in os.walk(output_directory):
          for f in files:
            if f[-4:].upper() == '.PNG' and f[:-4].isdigit():
              if cover_number < int(f[:-4]):
                cover_number = int(f[:-4])
        cover_filename = os.path.join(output_directory, str(cover_number + 1) + '.png')
        output = cStringIO.StringIO()
        coverpage.save(output, "PNG")
        coverpage_base64 = output.getvalue().encode("base64")
        output.close()
        coverpage.save(cover_filename, "PNG")

        # sequences
        sequences = ''
        for sequence in acbf_document.sequences:
          sequences = sequences + sequence[0] + ' (' + sequence[1] + '), '
        sequences = sequences[:-2]

        # publish-date 
        if acbf_document.publish_date_value != '':
          publish_date = acbf_document.publish_date_value
        else:
          publish_date = '9999-01-01'

        # languages
        languages = ''
        for language in acbf_document.languages:
          if language[1] == 'FALSE' and language[0] == '??':
            languages = languages + '??(no text layer), '
          elif language[1] == 'FALSE':
            languages = languages + language[0] + '(no text layer), '
          else:
            languages = languages + language[0] + ', '
        languages = languages[:-2]

        # image formats
        image_formats = []
        for page in acbf_document.pages:
          im_format = page.find("image").get("href")[-4:].upper().strip('.')
          if im_format not in image_formats:
            image_formats.append(im_format)
        im_formats = str(image_formats)[1:][:-1].replace("u'", "").replace("'", "")

        # clear library temp directory
        for root, dirs, files in os.walk(self._window.library_dir):
          for f in files:
            os.unlink(os.path.join(root, f))
          for d in dirs:
            shutil.rmtree(os.path.join(root, d))

        return cover_filename, acbf_document.book_title, publish_date, acbf_document.publisher, acbf_document.authors, acbf_document.genres, sequences, acbf_document.annotation, languages, acbf_document.characters, acbf_document.pages_total, acbf_document.license, acbf_document.has_frames, im_formats

    
class Library():

  def __init__(self):
      self.library_file_path = os.path.join(constants.CONFIG_DIR, 'library.xml');
      self.load_library()

  def create_new_tree(self):
      self.tree = xml.Element("library")

      version = xml.SubElement(self.tree, "version")
      version.text = constants.LIBRARY_VERSION

      library_info = xml.SubElement(self.tree, "library_info")

      #print(xml.tostring(self.tree, pretty_print=True))
      
  def load_library(self):
      if os.path.isfile(self.library_file_path):
        self.tree = xml.parse(source = self.library_file_path).getroot()
        if self.get_version() != constants.LIBRARY_VERSION:
           self.tree.find("version").text = constants.LIBRARY_VERSION
        if self.tree.find("library_info") == None:
          library_info = xml.SubElement(self.tree, "library_info")
      else:
        self.create_new_tree()
        f = open(self.library_file_path, 'w')
        f.write(xml.tostring(self.tree, pretty_print=True))
        f.close()

  def save_library(self):
      f = open(self.library_file_path, 'w')
      f.write(xml.tostring(self.tree, pretty_print=True))
      f.close()
      
  def delete_book(self, path):
      for book in self.tree.findall("book"):
        if book.get("path") == path:
          book.getparent().remove(book)
      return

  def get_version(self):
      if self.tree.find("version") != None:
        return self.tree.find("version").text
      else:
        return None

  def get_value(self, element, path):
      for book in self.tree.findall("book"):
        if book.get("path") == path:
          if book.find(element) != None:
            return book.find(element).text
          else:
            return None

  def set_value(self, element, value, path):
      for book in self.tree.findall("book"):
        if book.get("path") == path:
          if book.find(element) == None:
            new_element = xml.SubElement(book, element)
          book.find(element).text = value
          self.save_library()
      return

  def get_library_info_value(self, element):
      library_info = self.tree.find("library_info")
      if library_info.find(element) != None:
        return library_info.find(element).text
      else:
        return None

  def set_library_info_value(self, element, value):
      library_info = self.tree.find("library_info")
      if library_info.find(element) == None:
        new_element = xml.SubElement(library_info, element)
      library_info.find(element).text = value
      self.save_library()


def pil_to_pixbuf(PILImage, BGColor):
    """Return a pixbuf created from the PIL <image>."""

    bcolor = (int(gtk.gdk.color_parse(BGColor).red_float*255), int(gtk.gdk.color_parse(BGColor).green_float*255), int(gtk.gdk.color_parse(BGColor).blue_float*255))

    PILImage = PILImage.convert("RGBA")
    bg = Image.new("RGB", PILImage.size, bcolor)
    bg.paste(PILImage,PILImage)

    dummy_file = cStringIO.StringIO()
    bg.save(dummy_file, "ppm")
    contents = dummy_file.getvalue()
    dummy_file.close()

    loader = gtk.gdk.PixbufLoader("pnm")
    loader.write(contents, len(contents))
    pixbuf = loader.get_pixbuf()
    loader.close()
    return pixbuf

# function to retrieve text value from element without throwing exception
def get_element_text(element_tree, element):
    try:
      text_value = element_tree.find('{http://www.fictionbook-lib.org/xml/acbf/1.0}' + element).text
      if text_value is None:
        text_value = ''
    except:
      text_value = ''
    return text_value

def get_element_text2(element_tree, element):
    try:
      text_value = element_tree.find(element).text
      if text_value is None:
        text_value = ''
    except:
      text_value = ''
    return text_value

def get_element_text3(element_tree, element, lang):
    try:
      text_value = element_tree.find(element).text
      for i in element_tree.findall(element):
        if i.get("lang") == lang or (i.get("lang") == None and lang == "en"):
          text_value = i.text
      if text_value is None:
        text_value = ''
    except:
      text_value = ''
    return text_value

